# Defines errors and error codes

from contextlib import contextmanager

# Error codes
NO_DEVICE_TYPE = -1
MISSING_ARGUMENTS = -2
DEVICE_CONN_ERROR = -3
UNKNWON_DEVICE_TYPE = -4
INVALID_TX = -5
NO_PASSWORD = -6
BAD_ARGUMENT = -7
NOT_IMPLEMENTED = -8
UNAVAILABLE_ACTION = -9
DEVICE_ALREADY_INIT = -10
DEVICE_ALREADY_UNLOCKED = -11
DEVICE_NOT_READY = -12
UNKNOWN_ERROR = -13
ACTION_CANCELED = -14
DEVICE_BUSY = -15
NEED_TO_BE_ROOT = -16
HELP_TEXT = -17
DEVICE_NOT_INITIALIZED = -18

# Exceptions
class HWWError(Exception):
    def __init__(self, msg, code):
        Exception.__init__(self)
        self.code = code
        self.msg = msg

    def get_code(self):
        return self.code

    def get_msg(self):
        return self.msg

    def __str__(self):
        return self.msg

class NoPasswordError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, NO_PASSWORD)

class UnavailableActionError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, UNAVAILABLE_ACTION)

class DeviceAlreadyInitError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, DEVICE_ALREADY_INIT)

class DeviceNotReadyError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, DEVICE_NOT_READY)

class DeviceAlreadyUnlockedError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, DEVICE_ALREADY_UNLOCKED)

class UnknownDeviceError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, UNKNWON_DEVICE_TYPE)

class NotImplementedError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, NOT_IMPLEMENTED)

class PSBTSerializationError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, INVALID_TX)

class BadArgumentError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, BAD_ARGUMENT)

class DeviceFailureError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, UNKNOWN_ERROR)

class ActionCanceledError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, ACTION_CANCELED)

class DeviceConnectionError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, DEVICE_CONN_ERROR)

class DeviceBusyError(HWWError):
    def __init__(self, msg):
        HWWError.__init__(self, msg, DEVICE_BUSY)

@contextmanager
def handle_errors(msg=None, result=None, code=UNKNOWN_ERROR, debug=False):
    if result is None:
        result = {}

    if msg is None:
        msg = ""
    else:
        msg = msg + " "

    try:
        yield

    except HWWError as e:
        result['error'] = msg + e.get_msg()
        result['code'] = e.get_code()
    except Exception as e:
        result['error'] = msg + str(e)
        result['code'] = code
        if debug:
            import traceback
            traceback.print_exc()
    return result


common_err_msgs = {
    "enumerate": "Could not open client or get fingerprint information:"
}



from subprocess import check_call, CalledProcessError, DEVNULL
from .errors import NEED_TO_BE_ROOT
from shutil import copy
from os import path, listdir, getlogin, geteuid

class UDevInstaller(object):
    @staticmethod
    def install(source, location):
        try:
            udev_installer = UDevInstaller()
            udev_installer.copy_udev_rule_files(source, location)
            udev_installer.trigger()
            udev_installer.reload_rules()
            udev_installer.add_user_plugdev_group()
        except CalledProcessError:
            if geteuid() != 0:
                return {'error': 'Need to be root.', 'code': NEED_TO_BE_ROOT}
            raise
        return {"success": True}

    def __init__(self):
        self._udevadm = '/sbin/udevadm'
        self._groupadd = '/usr/sbin/groupadd'
        self._usermod = '/usr/sbin/usermod'

    def _execute(self, command, *args):
        command = [command] + list(args)
        check_call(command, stderr=DEVNULL, stdout=DEVNULL)

    def trigger(self):
        self._execute(self._udevadm, 'trigger')

    def reload_rules(self):
        self._execute(self._udevadm, 'control', '--reload-rules')

    def add_user_plugdev_group(self):
        self._create_group('plugdev')
        self._add_user_to_group(getlogin(), 'plugdev')

    def _create_group(self, name):
        try:
            self._execute(self._groupadd, name)
        except CalledProcessError as e:
            if e.returncode != 9: # group already exists
                raise

    def _add_user_to_group(self, user, group):
        self._execute(self._usermod, '-aG', group, user)

    def copy_udev_rule_files(self, source, location):
        src_dir_path = source
        for rules_file_name in listdir(_resource_path(src_dir_path)):
            if '.rules' in rules_file_name:
                rules_file_path = _resource_path(path.join(src_dir_path, rules_file_name))
                copy(rules_file_path, location)

def _resource_path(relative_path):
    return path.join(path.dirname(__file__), relative_path)



import re

# From: https://github.com/bitcoin/bitcoin/blob/master/src/script/descriptor.cpp

def PolyMod(c, val):
    c0 = c >> 35
    c = ((c & 0x7ffffffff) << 5) ^ val
    if (c0 & 1):
        c ^= 0xf5dee51989
    if (c0 & 2):
        c ^= 0xa9fdca3312
    if (c0 & 4):
        c ^= 0x1bab10e32d
    if (c0 & 8):
        c ^= 0x3706b1677a
    if (c0 & 16):
        c ^= 0x644d626ffd
    return c

def DescriptorChecksum(desc):
    INPUT_CHARSET = "0123456789()[],'/*abcdefgh@:$%{}IJKLMNOPQRSTUVWXYZ&+-.;<=>?!^_|~ijklmnopqrstuvwxyzABCDEFGH`#\"\\ "
    CHECKSUM_CHARSET = "qpzry9x8gf2tvdw0s3jn54khce6mua7l"

    c = 1
    cls = 0
    clscount = 0
    for ch in desc:
        pos = INPUT_CHARSET.find(ch)
        if pos == -1:
            return ""
        c = PolyMod(c, pos & 31)
        cls = cls * 3 + (pos >> 5)
        clscount += 1
        if clscount == 3:
            c = PolyMod(c, cls)
            cls = 0
            clscount = 0
    if clscount > 0:
        c = PolyMod(c, cls)
    for j in range(0, 8):
        c = PolyMod(c, 0)
    c ^= 1

    ret = [None] * 8
    for j in range(0, 8):
        ret[j] = CHECKSUM_CHARSET[(c >> (5 * (7 - j))) & 31]
    return ''.join(ret)

def AddChecksum(desc):
    return desc + "#" + DescriptorChecksum(desc)

class Descriptor:
    def __init__(
        self,
        origin_fingerprint,
        origin_path,
        base_key,
        path_suffix,
        testnet,
        sh_wpkh=None,
        wpkh=None,
        sh=None,
        sh_wsh=None,
        wsh=None,
        multisig_M=None,
        multisig_N=None
    ):
        self.origin_fingerprint = origin_fingerprint
        self.origin_path = origin_path
        self.path_suffix = path_suffix
        self.base_key = base_key
        self.testnet = testnet
        self.sh_wpkh = sh_wpkh
        self.wpkh = wpkh
        self.sh = sh
        self.sh_wsh = sh_wsh
        self.wsh = wsh
        self.multisig_M = multisig_M
        self.multisig_N = multisig_N
        self.m_path = None

        if origin_path and not isinstance(origin_path, list):
            self.m_path_base = "m" + origin_path
            self.m_path = "m" + origin_path + (path_suffix or "")

    @classmethod
    def parse(cls, desc, testnet=False):
        sh_wpkh = None
        wpkh = None
        sh = None
        sh_wsh = None
        wsh = None
        origin_fingerprint = None
        origin_path = None
        base_key_and_path_match = None
        base_key = None
        path_suffix = None
        multisig_M = None
        multisig_N = None

        # Check the checksum
        check_split = desc.split('#')
        if len(check_split) > 2:
            return None
        if len(check_split) == 2:
            if len(check_split[1]) != 8:
                return None
            checksum = DescriptorChecksum(check_split[0])
            if not checksum.strip():
                return None
            if checksum != check_split[1]:
                return None
        desc = check_split[0]

        if desc.startswith("sh(wpkh("):
            sh_wpkh = True
        elif desc.startswith("wpkh("):
            wpkh = True
        elif desc.startswith("sh(wsh("):
            sh_wsh = True
        elif desc.startswith("wsh("):
            wsh = True
        elif desc.startswith("sh("):
            sh = True

        if sh or sh_wsh or wsh:
            if 'multi(' not in desc:
                # only multisig scripts are supported
                return None
            # get the list of keys only
            keys = desc.split(',', 1)[1].split(')', 1)[0].split(',')
            if 'sortedmulti' in desc:
                keys.sort(key=lambda x: x if ']' not in x else x.split(']')[1])
            multisig_M = desc.split(',')[0].split('(')[-1]
            multisig_N = len(keys)
        else:
            keys = [desc.split('(')[-1].split(')', 1)[0]]

        descriptors = []
        for key in keys:
            origin_match = re.search(r"\[(.*)\]", key)
            if origin_match:
                origin = origin_match.group(1)
                match = re.search(r"^([0-9a-fA-F]{8})(\/.*)", origin)
                if match:
                    origin_fingerprint = match.group(1)
                    origin_path = match.group(2)
                    # Replace h with '
                    origin_path = origin_path.replace('h', '\'')

                base_key_and_path_match = re.search(r"\[.*\](\w+)([\d'\/\*]*)", key)
            else:
                base_key_and_path_match = re.search(r"(\w+)([\d'\/\*]*)", key)

            if base_key_and_path_match:
                base_key = base_key_and_path_match.group(1)
                path_suffix = base_key_and_path_match.group(2)
                if path_suffix == '':
                    path_suffix = None
            else:
                if origin_match is None:
                    return None

            descriptors.append(cls(origin_fingerprint, origin_path, base_key, path_suffix, testnet, sh_wpkh, wpkh, sh, sh_wsh, wsh))
        if len(descriptors) == 1:
            return descriptors[0]
        else:
            # for multisig scripts save as lists all keypaths fields
            return cls(
                [descriptor.origin_fingerprint for descriptor in descriptors],
                [descriptor.origin_path for descriptor in descriptors],
                [descriptor.base_key for descriptor in descriptors],
                [descriptor.path_suffix for descriptor in descriptors],
                testnet,
                sh_wpkh,
                wpkh,
                sh,
                sh_wsh,
                wsh,
                multisig_M,
                multisig_N
            )

    def serialize(self):
        descriptor_open = 'pkh('
        descriptor_close = ')'
        origin = ''
        path_suffix = ''

        if self.wpkh:
            descriptor_open = 'wpkh('
        elif self.sh_wpkh:
            descriptor_open = 'sh(wpkh('
            descriptor_close = '))'
        elif self.sh or self.sh_wsh or self.wsh:
            # serialize multisig descriptor is not supported yet.
            return None

        if self.origin_fingerprint and self.origin_path:
            origin = '[' + self.origin_fingerprint + self.origin_path + ']'

        if self.path_suffix:
            path_suffix = self.path_suffix

        return AddChecksum(descriptor_open + origin + self.base_key + path_suffix + descriptor_close)



# KeepKey interaction script

from ..errors import (
    DEVICE_NOT_INITIALIZED,
    DeviceNotReadyError,
    common_err_msgs,
    handle_errors,
)
from .trezorlib.transport import (
    enumerate_devices,
    KEEPKEY_VENDOR_IDS,
)
from .trezor import TrezorClient

py_enumerate = enumerate # Need to use the enumerate built-in but there's another function already named that

class KeepkeyClient(TrezorClient):
    def __init__(self, path, password='', expert=False):
        super(KeepkeyClient, self).__init__(path, password, expert)
        self.type = 'Keepkey'

def enumerate(password=''):
    results = []
    for dev in enumerate_devices():
        # enumerate_devices filters to Trezors and Keepkeys.
        # Only allow Keepkeys and unknowns. Unknown devices will reach the check for vendor later
        if dev.get_usb_vendor_id() not in KEEPKEY_VENDOR_IDS | {-1}:
            continue
        d_data = {}

        d_data['type'] = 'keepkey'
        d_data['model'] = 'keepkey'
        d_data['path'] = dev.get_path()

        client = None

        with handle_errors(common_err_msgs["enumerate"], d_data):
            client = KeepkeyClient(d_data['path'], password)
            client.client.init_device()
            if 'keepkey' not in client.client.features.vendor:
                continue

            if d_data['path'] == 'udp:127.0.0.1:21324':
                d_data['model'] += '_simulator'

            d_data['needs_pin_sent'] = client.client.features.pin_protection and not client.client.features.pin_cached
            d_data['needs_passphrase_sent'] = client.client.features.passphrase_protection # always need the passphrase sent for Keepkey if it has passphrase protection enabled
            if d_data['needs_pin_sent']:
                raise DeviceNotReadyError('Keepkey is locked. Unlock by using \'promptpin\' and then \'sendpin\'.')
            if d_data['needs_passphrase_sent'] and not password:
                raise DeviceNotReadyError("Passphrase needs to be specified before the fingerprint information can be retrieved")
            if client.client.features.initialized:
                d_data['fingerprint'] = client.get_master_fingerprint_hex()
                d_data['needs_passphrase_sent'] = False # Passphrase is always needed for the above to have worked, so it's already sent
            else:
                d_data['error'] = 'Not initialized'
                d_data['code'] = DEVICE_NOT_INITIALIZED

        if client:
            client.close()

        results.append(d_data)
    return results



# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class Deprecated_PassphraseStateRequest(p.MessageType):
    MESSAGE_WIRE_TYPE = 77

    def __init__(
        self,
        _state: bytes = None,
    ) -> None:
        self._state = _state

    @classmethod
    def get_fields(cls):
        return {
            1: ('_state', p.BytesType, 0),
        }



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['hwilib',
 'hwilib.devices',
 'hwilib.devices.btchip',
 'hwilib.devices.ckcc',
 'hwilib.devices.trezorlib',
 'hwilib.devices.trezorlib.messages',
 'hwilib.devices.trezorlib.transport']

package_data = \
{'': ['*'],
 'hwilib': ['udev/*',
            'ui/displayaddressdialog.ui',
            'ui/displayaddressdialog.ui',
            'ui/displayaddressdialog.ui',
            'ui/displayaddressdialog.ui',
            'ui/displayaddressdialog.ui',
            'ui/displayaddressdialog.ui',
            'ui/displayaddressdialog.ui',
            'ui/displayaddressdialog.ui',
            'ui/displayaddressdialog.ui',
            'ui/getkeypooloptionsdialog.ui',
            'ui/getkeypooloptionsdialog.ui',
            'ui/getkeypooloptionsdialog.ui',
            'ui/getkeypooloptionsdialog.ui',
            'ui/getkeypooloptionsdialog.ui',
            'ui/getkeypooloptionsdialog.ui',
            'ui/getkeypooloptionsdialog.ui',
            'ui/getkeypooloptionsdialog.ui',
            'ui/getkeypooloptionsdialog.ui',
            'ui/getxpubdialog.ui',
            'ui/getxpubdialog.ui',
            'ui/getxpubdialog.ui',
            'ui/getxpubdialog.ui',
            'ui/getxpubdialog.ui',
            'ui/getxpubdialog.ui',
            'ui/getxpubdialog.ui',
            'ui/getxpubdialog.ui',
            'ui/getxpubdialog.ui',
            'ui/hwiqt.pyproject',
            'ui/hwiqt.pyproject',
            'ui/hwiqt.pyproject',
            'ui/hwiqt.pyproject',
            'ui/hwiqt.pyproject',
            'ui/hwiqt.pyproject',
            'ui/hwiqt.pyproject',
            'ui/hwiqt.pyproject',
            'ui/hwiqt.pyproject',
            'ui/mainwindow.ui',
            'ui/mainwindow.ui',
            'ui/mainwindow.ui',
            'ui/mainwindow.ui',
            'ui/mainwindow.ui',
            'ui/mainwindow.ui',
            'ui/mainwindow.ui',
            'ui/mainwindow.ui',
            'ui/mainwindow.ui',
            'ui/sendpindialog.ui',
            'ui/sendpindialog.ui',
            'ui/sendpindialog.ui',
            'ui/sendpindialog.ui',
            'ui/sendpindialog.ui',
            'ui/sendpindialog.ui',
            'ui/sendpindialog.ui',
            'ui/sendpindialog.ui',
            'ui/sendpindialog.ui',
            'ui/setpassphrasedialog.ui',
            'ui/setpassphrasedialog.ui',
            'ui/setpassphrasedialog.ui',
            'ui/setpassphrasedialog.ui',
            'ui/setpassphrasedialog.ui',
            'ui/setpassphrasedialog.ui',
            'ui/setpassphrasedialog.ui',
            'ui/setpassphrasedialog.ui',
            'ui/setpassphrasedialog.ui',
            'ui/signmessagedialog.ui',
            'ui/signmessagedialog.ui',
            'ui/signmessagedialog.ui',
            'ui/signmessagedialog.ui',
            'ui/signmessagedialog.ui',
            'ui/signmessagedialog.ui',
            'ui/signmessagedialog.ui',
            'ui/signmessagedialog.ui',
            'ui/signmessagedialog.ui',
            'ui/signpsbtdialog.ui',
            'ui/signpsbtdialog.ui',
            'ui/signpsbtdialog.ui',
            'ui/signpsbtdialog.ui',
            'ui/signpsbtdialog.ui',
            'ui/signpsbtdialog.ui',
            'ui/signpsbtdialog.ui',
            'ui/signpsbtdialog.ui',
            'ui/signpsbtdialog.ui']}

modules = \
['hwi', 'hwi-qt']
install_requires = \
['ecdsa>=0.13.0,<0.14.0',
 'hidapi>=0.7.99,<0.8.0',
 'libusb1>=1.7,<2.0',
 'mnemonic>=0.18.0,<0.19.0',
 'pyaes>=1.6,<2.0',
 'typing-extensions>=3.7,<4.0']

extras_require = \
{'qt': ['pyside2>=5.14.0,<6.0.0']}

entry_points = \
{'console_scripts': ['hwi = hwilib.cli:main', 'hwi-qt = hwilib.gui:main']}

setup_kwargs = {
    'name': 'hwi',
    'version': '1.1.2',
    'description': 'A library for working with Bitcoin hardware wallets',
    'long_description': "# Bitcoin Hardware Wallet Interface\n\n[![Build Status](https://travis-ci.org/bitcoin-core/HWI.svg?branch=master)](https://travis-ci.org/bitcoin-core/HWI)\n\nThe Bitcoin Hardware Wallet Interface is a Python library and command line tool for interacting with hardware wallets.\nIt provides a standard way for software to work with hardware wallets without needing to implement device specific drivers.\nPython software can use the provided library (`hwilib`). Software in other languages can execute the `hwi` tool.\n\n## Prerequisites\n\nPython 3 is required. The libraries and [udev rules](hwilib/udev/README.md) for each device must also be installed. Some libraries will need to be installed\n\nFor Ubuntu/Debian:\n```\nsudo apt install libusb-1.0-0-dev libudev-dev python3-dev\n```\n\nFor Centos:\n```\nsudo yum -y install python3-devel libusbx-devel systemd-devel\n```\n\nFor macOS:\n```\nbrew install libusb\n```\n\n## Install\n\n```\ngit clone https://github.com/bitcoin-core/HWI.git\ncd HWI\npoetry install # or 'pip3 install .' or 'python3 setup.py install'\n```\n\nThis project uses the [Poetry](https://github.com/sdispater/poetry) dependency manager. HWI and its dependencies can be installed via poetry by executing the following in the root source directory:\n\n```\npoetry install\n```\n\nPip can also be used to automatically install HWI and its dependencies using the `setup.py` file (which is usually in sync with `pyproject.toml`):\n\n```\npip3 install .\n```\n\nThe `setup.py` file can be used to install HWI and its dependencies so long as `setuptools` is also installed:\n\n```\npip3 install -U setuptools\npython3 setup.py install\n```\n\n## Dependencies\n\nSee `pyproject.toml` for all dependencies. Dependencies under `[tool.poetry.dependecies]` are user dependencies, and `[tool.poetry.dev-dependencies]` for development based dependencies. These dependencies will be installed with any of the three above installation methods.\n\n## Usage\n\nTo use, first enumerate all devices and find the one that you want to use with\n\n```\n./hwi.py enumerate\n```\n\nOnce the device type and device path is known, issue commands to it like so:\n\n```\n./hwi.py -t <type> -d <path> <command> <command args>\n```\n\nAll output will be in JSON form and sent to `stdout`.\nAdditional information or prompts will be sent to `stderr` and will not necessarily be in JSON.\nThis additional information is for debugging purposes.\n\nTo see a complete list of available commands and global parameters, run\n`./hwi.py --help`.  To see options specific to a particular command,\npass the `--help` parameter after the command name; for example:\n\n```\n./hwi.py getdescriptors --help\n```\n\n## Device Support\n\nThe below table lists what devices and features are supported for each device.\n\nPlease also see [docs](docs/) for additional information about each device.\n\n| Feature \\ Device | Ledger Nano X | Ledger Nano S | Trezor One | Trezor Model T | Digital BitBox | KeepKey | Coldcard |\n|:---:|:---:|:---:|:---:|:---:|:---:|:---:|:---:|\n| Support Planned | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| Implemented | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| xpub retrieval | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| Message Signing | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| Device Setup | N/A | N/A | Yes | Yes | Yes | Yes | N/A |\n| Device Wipe | N/A | N/A | Yes | Yes | Yes | Yes | N/A |\n| Device Recovery | N/A | N/A | Yes | Yes | N/A | Yes | N/A |\n| Device Backup | N/A | N/A | N/A | N/A | Yes | N/A | Yes |\n| P2PKH Inputs | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| P2SH-P2WPKH Inputs | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| P2WPKH Inputs | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| P2SH Multisig Inputs | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| P2SH-P2WSH Multisig Inputs | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| P2WSH Multisig Inputs | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| Bare Multisig Inputs | Yes | Yes | N/A | N/A | Yes | N/A | N/A |\n| Arbitrary scriptPubKey Inputs | Yes | Yes | N/A | N/A | Yes | N/A | N/A |\n| Arbitrary redeemScript Inputs | Yes | Yes | N/A | N/A | Yes | N/A | N/A |\n| Arbitrary witnessScript Inputs | Yes | Yes | N/A | N/A | Yes | N/A | N/A |\n| Non-wallet inputs | Yes | Yes | Yes | Yes | Yes | Yes | Yes |\n| Mixed Segwit and Non-Segwit Inputs | N/A | N/A | Yes | N/A | Yes | Yes | Yes |\n| Display on device screen | Yes | Yes | Yes | Yes | N/A | Yes | Yes |\n\n## Using with Bitcoin Core\n\nSee [Using Bitcoin Core with Hardware Wallets](docs/bitcoin-core-usage.md).\n\n## License\n\nThis project is available under the MIT License, Copyright Andrew Chow.\n",
    'author': 'Andrew Chow',
    'author_email': 'andrew@achow101.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bitcoin-core/HWI',
    'packages': packages,
    'package_data': package_data,
    'py_modules': modules,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<3.9',
}


setup(**setup_kwargs)



#! /usr/bin/env python3

# Hardware wallet interaction script

import importlib
import platform

from .serializations import PSBT
from .base58 import xpub_to_pub_hex
from .errors import (
    UnknownDeviceError,
    BAD_ARGUMENT,
    NOT_IMPLEMENTED,
)
from .descriptor import Descriptor
from .devices import __all__ as all_devs
from enum import Enum

class AddressType(Enum):
    PKH = 1
    WPKH = 2
    SH_WPKH = 3

# Get the client for the device
def get_client(device_type, device_path, password='', expert=False):
    device_type = device_type.split('_')[0]
    class_name = device_type.capitalize()
    module = device_type.lower()

    client = None
    try:
        imported_dev = importlib.import_module('.devices.' + module, __package__)
        client_constructor = getattr(imported_dev, class_name + 'Client')
        client = client_constructor(device_path, password, expert)
    except ImportError:
        if client:
            client.close()
        raise UnknownDeviceError('Unknown device type specified')

    return client

# Get a list of all available hardware wallets
def enumerate(password=''):
    result = []

    for module in all_devs:
        try:
            imported_dev = importlib.import_module('.devices.' + module, __package__)
            result.extend(imported_dev.enumerate(password))
        except ImportError:
            pass # Ignore ImportErrors, the user may not have all device dependencies installed
    return result

# Fingerprint or device type required
def find_device(password='', device_type=None, fingerprint=None, expert=False):
    devices = enumerate(password)
    for d in devices:
        if device_type is not None and d['type'] != device_type and d['model'] != device_type:
            continue
        client = None
        try:
            client = get_client(d['type'], d['path'], password, expert)

            master_fpr = d.get('fingerprint', None)
            if master_fpr is None:
                master_fpr = client.get_master_fingerprint_hex()

            if fingerprint and master_fpr != fingerprint:
                client.close()
                continue
            return client
        except:
            if client:
                client.close()
            pass # Ignore things we wouldn't get fingerprints for
    return None

def getmasterxpub(client):
    return client.get_master_xpub()

def signtx(client, psbt):
    # Deserialize the transaction
    tx = PSBT()
    tx.deserialize(psbt)
    return client.sign_tx(tx)

def getxpub(client, path):
    return client.get_pubkey_at_path(path)

def signmessage(client, message, path):
    return client.sign_message(message, path)

def getkeypool_inner(client, path, start, end, internal=False, keypool=True, account=0, addr_type=AddressType.WPKH):

    try:
        master_fpr = client.get_master_fingerprint_hex()
    except NotImplementedError as e:
        return {'error': str(e), 'code': NOT_IMPLEMENTED}

    desc = getdescriptor(client, master_fpr, client.is_testnet, path, internal, addr_type, account, start, end)

    if not isinstance(desc, Descriptor):
        return desc

    this_import = {}

    this_import['desc'] = desc.serialize()
    this_import['range'] = [start, end]
    this_import['timestamp'] = 'now'
    this_import['internal'] = internal
    this_import['keypool'] = keypool
    this_import['active'] = keypool
    this_import['watchonly'] = True
    return [this_import]

def getdescriptor(client, master_fpr, testnet=False, path=None, internal=False, addr_type=AddressType.WPKH, account=0, start=None, end=None):
    testnet = client.is_testnet

    is_wpkh = addr_type is AddressType.WPKH
    is_sh_wpkh = addr_type is AddressType.SH_WPKH

    if not path:
        # Master key:
        path = "m/"

        # Purpose
        if is_wpkh:
            path += "84'/"
        elif is_sh_wpkh:
            path += "49'/"
        else:
            assert addr_type == AddressType.PKH
            path += "44'/"

        # Coin type
        if testnet:
            path += "1'/"
        else:
            path += "0'/"

        # Account
        path += str(account) + '\'/'

        # Receive or change
        if internal:
            path += "1/*"
        else:
            path += "0/*"
    else:
        if path[0] != "m":
            return {'error': 'Path must start with m/', 'code': BAD_ARGUMENT}
        if path[-1] != "*":
            return {'error': 'Path must end with /*', 'code': BAD_ARGUMENT}

    # Find the last hardened derivation:
    path = path.replace('\'', 'h')
    path_suffix = ''
    for component in path.split("/")[::-1]:
        if component[-1] == 'h' or component[-1] == 'm':
            break
        path_suffix = '/' + component + path_suffix
    path_base = path.rsplit(path_suffix)[0]

    # Get the key at the base
    if client.xpub_cache.get(path_base) is None:
        client.xpub_cache[path_base] = client.get_pubkey_at_path(path_base)['xpub']

    return Descriptor(master_fpr, path_base.replace('m', ''), client.xpub_cache.get(path_base), path_suffix, client.is_testnet, is_sh_wpkh, is_wpkh)

def getkeypool(client, path, start, end, internal=False, keypool=True, account=0, sh_wpkh=False, wpkh=True, addr_all=False):

    if sh_wpkh:
        addr_types = [AddressType.SH_WPKH]
    elif wpkh:
        addr_types = [AddressType.WPKH]
    elif addr_all:
        addr_types = list(AddressType)
    else:
        addr_types = [AddressType.PKH]

    # When no specific path or internal-ness is specified, create standard types
    chains = []
    if path is None and not internal:
        for addr_type in addr_types:
            for internal_addr in [False, True]:
                chains = chains + getkeypool_inner(client, None, start, end, internal_addr, keypool, account, addr_type)

        # Report the first error we encounter
        for chain in chains:
            if 'error' in chain:
                return chain
        # No errors, return pair
        return chains
    else:
        assert len(addr_types) == 1
        return getkeypool_inner(client, path, start, end, internal, keypool, account, addr_types[0])


def getdescriptors(client, account=0):
    try:
        master_fpr = client.get_master_fingerprint_hex()
    except NotImplementedError as e:
        return {'error': str(e), 'code': NOT_IMPLEMENTED}

    result = {}

    for internal in [False, True]:
        descriptors = []
        desc1 = getdescriptor(client, master_fpr=master_fpr, testnet=client.is_testnet, internal=internal, addr_type=AddressType.PKH, account=account)
        desc2 = getdescriptor(client, master_fpr=master_fpr, testnet=client.is_testnet, internal=internal, addr_type=AddressType.SH_WPKH, account=account)
        desc3 = getdescriptor(client, master_fpr=master_fpr, testnet=client.is_testnet, internal=internal, addr_type=AddressType.WPKH, account=account)
        for desc in [desc1, desc2, desc3]:
            if not isinstance(desc, Descriptor):
                return desc
            descriptors.append(desc.serialize())
        if internal:
            result["internal"] = descriptors
        else:
            result["receive"] = descriptors

    return result

def displayaddress(client, path=None, desc=None, sh_wpkh=False, wpkh=False, redeem_script=None):
    if path is not None:
        if sh_wpkh and wpkh:
            return {'error': 'Both `--wpkh` and `--sh_wpkh` can not be selected at the same time.', 'code': BAD_ARGUMENT}
        return client.display_address(path, sh_wpkh, wpkh, redeem_script=redeem_script)
    elif desc is not None:
        if sh_wpkh or wpkh:
            return {'error': ' `--wpkh` and `--sh_wpkh` can not be combined with --desc', 'code': BAD_ARGUMENT}
        if redeem_script:
            return {'error': ' `--redeem_script` can not be combined with --desc', 'code': BAD_ARGUMENT}
        descriptor = Descriptor.parse(desc, client.is_testnet)
        if descriptor is None:
            return {'error': 'Unable to parse descriptor: ' + desc, 'code': BAD_ARGUMENT}
        if descriptor.sh or descriptor.sh_wsh or descriptor.wsh:
            path = ''
            redeem_script = format(80 + int(descriptor.multisig_M), 'x')
            for i in range(0, descriptor.multisig_N):
                path += descriptor.origin_fingerprint[i] + descriptor.origin_path[i] + ','
                if not descriptor.path_suffix[i]:
                    redeem_script += '21' + descriptor.base_key[i]
                else:
                    return {'error': 'Multisig descriptor must include all pubkeys', 'code': BAD_ARGUMENT}
            path = path[0:-1]
            redeem_script += format(80 + descriptor.multisig_N, 'x') + 'ae'
            return client.display_address(path, descriptor.sh_wpkh or descriptor.sh_wsh, descriptor.wpkh or descriptor.wsh, redeem_script)
        if descriptor.m_path is None:
            return {'error': 'Descriptor missing origin info: ' + desc, 'code': BAD_ARGUMENT}
        if descriptor.origin_fingerprint != client.get_master_fingerprint_hex():
            return {'error': 'Descriptor fingerprint does not match device: ' + desc, 'code': BAD_ARGUMENT}
        xpub = client.get_pubkey_at_path(descriptor.m_path_base)['xpub']
        if descriptor.base_key != xpub and descriptor.base_key != xpub_to_pub_hex(xpub):
            return {'error': 'Key in descriptor does not match device: ' + desc, 'code': BAD_ARGUMENT}
        return client.display_address(descriptor.m_path, descriptor.sh_wpkh, descriptor.wpkh)

def setup_device(client, label='', backup_passphrase=''):
    return client.setup_device(label, backup_passphrase)

def wipe_device(client):
    return client.wipe_device()

def restore_device(client, label='', word_count=24):
    return client.restore_device(label, word_count)

def backup_device(client, label='', backup_passphrase=''):
    return client.backup_device(label, backup_passphrase)

def prompt_pin(client):
    return client.prompt_pin()

def send_pin(client, pin):
    return client.send_pin(pin)

def toggle_passphrase(client):
    return client.toggle_passphrase()

def install_udev_rules(source, location):
    if platform.system() == "Linux":
        from .udevinstaller import UDevInstaller
        return UDevInstaller.install(source, location)
    return {'error': 'udev rules are not needed on your platform', 'code': NOT_IMPLEMENTED}
